#include <stdio.h>
#include <stdlib.h>

extern "C" {
#define XMD_H
#define NEED_FAR_POINTERS
#include "jpeglib.h"
#include "jinclude.h"
};

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "mlibjpg.h"


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
    }
    return TRUE;
}


struct my_error_mgr {
	struct jpeg_error_mgr pub;	/* "public" fields */
};

typedef my_error_mgr *my_error_ptr;

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	my_error_ptr myerr = (my_error_ptr) cinfo->err;

	throw cinfo->err->msg_code;
}

//---------------------------------------------------------------------------
//  LoadJPG
//---------------------------------------------------------------------------
extern "C" MLIBJPG_API int LoadJPG(const char *filename, 
		int (*header_callback)(LoadJPGInfo *info, void *data), 
		void *(*read_callback)(int line_number, void *data),
		void *data)
{
	jpeg_decompress_struct cinfo;
	my_error_mgr jerr;
	FILE *infile;
	JSAMPARRAY buffer;
	int row_stride;
	unsigned char *rp,*wp;
	LoadJPGInfo info;
	unsigned x;
	int result = 1;

	if(header_callback==NULL || read_callback==NULL)
		return 0;

	infile = fopen(filename, "rb");
	if(infile == NULL)
		return 0;

try {

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, true);

	info.width = cinfo.image_width;
	info.height = cinfo.image_height;
	info.bpp = 0;
	info.num_palette = 0;
	info.palette = NULL;

	switch(cinfo.out_color_space){
	case JCS_GRAYSCALE:
		if(cinfo.num_components == 1){
			info.bpp = 8;
			info.num_palette = 256;
			info.palette = (RGBQUAD *)malloc(sizeof(RGBQUAD) * 256);
			for(x=0; x<256; x++){
				info.palette[x].rgbBlue = info.palette[x].rgbGreen = 
					info.palette[x].rgbRed = x;
				info.palette[x].rgbReserved = 0;
			}
		}
		break;
	case JCS_RGB:
		if(cinfo.num_components == 3)
			info.bpp = 24;
		break;
	case JCS_CMYK:
		if(cinfo.num_components == 4)
			info.bpp = 24;
		break;
	}

	if(info.bpp != 0){
		if(!header_callback(&info, data))
			info.bpp = 0;
	}

	if(info.palette != NULL)
		free(info.palette);

	if(info.bpp == 0)
	{
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		return 0;
	}
	
	jpeg_start_decompress(&cinfo);

	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	while(cinfo.output_scanline < cinfo.output_height)
	{
		wp = (unsigned char *)read_callback(cinfo.output_scanline, data);
		if(wp == NULL)
			break;

		jpeg_read_scanlines(&cinfo, buffer, 1);

		switch(cinfo.out_color_space){
		case JCS_GRAYSCALE:
			memcpy(wp, buffer[0], cinfo.output_width);
			break;
		case JCS_RGB:
			for(x=0,rp=buffer[0]; x<cinfo.output_width; x++,rp+=3){
				*wp++ = rp[2];
				*wp++ = rp[1];
				*wp++ = rp[0];
			}
			break;
		case JCS_CMYK:
			for(x=0,rp=buffer[0]; x<cinfo.output_width; x++,rp+=4){
				*wp++ = (rp[2] * rp[3]) / 255;
				*wp++ = (rp[1] * rp[3]) / 255;
				*wp++ = (rp[0] * rp[3]) / 255;
			}
		}
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

} catch(int /*error_code*/) {
	result = 0;
}

	fclose(infile);

	return result;
}	//LoadJPG()


//---------------------------------------------------------------------------
//  LoadJPG
//---------------------------------------------------------------------------

/* Expanded data source object for stdio input */

typedef struct {
	struct jpeg_source_mgr pub;	/* public fields */

	void *stream;
	int size;
	int count;
	boolean start_of_file;	/* have we gotten any data yet? */
} mlibjpg_source_mgr;

typedef mlibjpg_source_mgr * mlibjpg_src_ptr;

METHODDEF(void)
init_source (j_decompress_ptr cinfo)
{
	mlibjpg_src_ptr src = (mlibjpg_src_ptr) cinfo->src;

  /* We reset the empty-input-file flag for each image,
   * but we don't clear the input buffer.
   * This is correct behavior for reading a series of images from one source.
   */
	src->start_of_file = TRUE;
}

METHODDEF(boolean)
fill_input_buffer (j_decompress_ptr cinfo)
{
	mlibjpg_src_ptr src = (mlibjpg_src_ptr) cinfo->src;

	src->pub.next_input_byte = (unsigned char *)src->stream;
	src->pub.bytes_in_buffer = src->size;
	src->start_of_file = FALSE;

	return TRUE;
}


METHODDEF(void)
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
	mlibjpg_src_ptr src = (mlibjpg_src_ptr) cinfo->src;

	if (num_bytes > 0) {
		while (num_bytes > (long) src->pub.bytes_in_buffer)
		{
			num_bytes -= (long) src->pub.bytes_in_buffer;
			(void) fill_input_buffer(cinfo);
			/* note we assume that fill_input_buffer will never return FALSE,
			 * so suspension need not be handled.
			 */
		}
		src->pub.next_input_byte += (size_t) num_bytes;
		src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
}


METHODDEF(void)
term_source (j_decompress_ptr cinfo)
{
	/* no work necessary here */
}


GLOBAL(void)
jpeg_memory_src(j_decompress_ptr cinfo, void *stream, int stream_size)
{
	mlibjpg_src_ptr src;

	if (cinfo->src == NULL) {	/* first time for this JPEG object? */
		cinfo->src = (struct jpeg_source_mgr *)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				SIZEOF(mlibjpg_source_mgr));
	}

	src = (mlibjpg_src_ptr) cinfo->src;
	src->pub.init_source = init_source;
	src->pub.fill_input_buffer = fill_input_buffer;
	src->pub.skip_input_data = skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = term_source;
	src->stream = stream;
	src->size = stream_size;
	src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	src->pub.next_input_byte = NULL; /* until buffer loaded */
}

extern "C" MLIBJPG_API int LoadJPGFromMemory(void *memory, int memory_size,
		int (*header_callback)(LoadJPGInfo *info, void *data), 
		void *(*read_callback)(int line_number, void *data),
		void *data)
{
	jpeg_decompress_struct cinfo;
	my_error_mgr jerr;
	JSAMPARRAY buffer;
	int row_stride;
	unsigned char *rp,*wp;
	LoadJPGInfo info;
	unsigned x;
	int result = 1;

	if(header_callback==NULL || read_callback==NULL)
		return 0;

try {

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	jpeg_create_decompress(&cinfo);
	
	//jpeg_stdio_src(&cinfo, infile);
	jpeg_memory_src(&cinfo, memory, memory_size);
	
	jpeg_read_header(&cinfo, true);

	info.width = cinfo.image_width;
	info.height = cinfo.image_height;
	info.bpp = 0;
	info.num_palette = 0;
	info.palette = NULL;

	switch(cinfo.out_color_space){
	case JCS_GRAYSCALE:
		if(cinfo.num_components == 1){
			info.bpp = 8;
			info.num_palette = 256;
			info.palette = (RGBQUAD *)malloc(sizeof(RGBQUAD) * 256);
			for(x=0; x<256; x++){
				info.palette[x].rgbBlue = info.palette[x].rgbGreen = 
					info.palette[x].rgbRed = x;
				info.palette[x].rgbReserved = 0;
			}
		}
		break;
	case JCS_RGB:
		if(cinfo.num_components == 3)
			info.bpp = 24;
		break;
	case JCS_CMYK:
		if(cinfo.num_components == 4)
			info.bpp = 24;
		break;
	}

	if(info.bpp != 0){
		if(!header_callback(&info, data))
			info.bpp = 0;
	}

	if(info.palette != NULL)
		free(info.palette);

	if(info.bpp == 0)
	{
		jpeg_destroy_decompress(&cinfo);
		return 0;
	}
	
	jpeg_start_decompress(&cinfo);

	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	while(cinfo.output_scanline < cinfo.output_height)
	{
		wp = (unsigned char *)read_callback(cinfo.output_scanline, data);
		if(wp == NULL)
			break;

		jpeg_read_scanlines(&cinfo, buffer, 1);

		switch(cinfo.out_color_space){
		case JCS_GRAYSCALE:
			memcpy(wp, buffer[0], cinfo.output_width);
			break;
		case JCS_RGB:
			for(x=0,rp=buffer[0]; x<cinfo.output_width; x++,rp+=3){
				*wp++ = rp[2];
				*wp++ = rp[1];
				*wp++ = rp[0];
			}
			break;
		case JCS_CMYK:
			for(x=0,rp=buffer[0]; x<cinfo.output_width; x++,rp+=4){
				*wp++ = (rp[2] * rp[3]) / 255;
				*wp++ = (rp[1] * rp[3]) / 255;
				*wp++ = (rp[0] * rp[3]) / 255;
			}
		}
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

} catch(int /*error_code*/) {
	result = 0;
}

	return result;
}	//LoadJPGFromMemory()


//---------------------------------------------------------------------------
//  SaveJPG
//---------------------------------------------------------------------------
extern "C" MLIBJPG_API int SaveJPG(const char *filename, 
		SaveJPGInfo *info,
		void *(*write_callback)(int line_number, void *data),
		void *data)
{
	jpeg_compress_struct cinfo;
	my_error_mgr jerr;
	JSAMPARRAY buffer;
	int row_stride;
	FILE *outfile;
	int result = 1;
	unsigned char *rp, *wp;
	int x,y;

	if(write_callback == NULL)
		return 0;

	if(info->bpp != 24)
		return 0;

	/* Open the output file. */
    outfile = fopen(filename, "wb");
	if(outfile == NULL) {
		return 0;
    }

try {

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	jpeg_create_compress(&cinfo);

	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = info->width; 	/* image width and height, in pixels */
	cinfo.image_height = info->height;
	cinfo.input_components = 3;		/* # of color components per pixel */
	cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
	jpeg_set_defaults(&cinfo);

	if(info->quality > 0)
		jpeg_set_quality(&cinfo, info->quality, info->force_baseline);

	if(info->progressive)
		jpeg_simple_progression(&cinfo);

	/* Now that we know input colorspace, fix colorspace-dependent defaults */
	//jpeg_default_colorspace(&cinfo);

	/* Start compressor */
	jpeg_start_compress(&cinfo, TRUE);

	/* Process data */
	row_stride = cinfo.image_width * cinfo.input_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
	for(y=0; y<info->height; y++)
	{
		rp = (unsigned char *)write_callback(y, data);
		if(rp == NULL) break;
		wp = buffer[0];
		for(x=0; x<info->width; x++,rp+=3){
			*wp++ = rp[2];
			*wp++ = rp[1];
			*wp++ = rp[0];
		}
		jpeg_write_scanlines(&cinfo, buffer, 1);
	}

	/* Finish compression and release memory */
	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

} catch(int /*error_code*/){
	result = 0;
}

	fclose(outfile);

	return result;
}	//SaveJPG()


//---------------------------------------------------------------------------
//  SaveJPGToMemory
//---------------------------------------------------------------------------

#define STREAM_SIZE   65536

typedef struct {
	struct jpeg_destination_mgr pub; /* public fields */
	void **stream;		/* target stream */
	int stream_count;
	int totalsize;
} mlibjpg_destination_mgr;

typedef mlibjpg_destination_mgr * mlibjpg_dest_ptr;

METHODDEF(void)
init_destination (j_compress_ptr cinfo)
{
	mlibjpg_dest_ptr dest = (mlibjpg_dest_ptr) cinfo->dest;
	
	dest->stream = (void **)malloc(sizeof(void*));
	dest->stream[0] = malloc(STREAM_SIZE);

	dest->pub.next_output_byte = (unsigned char *)dest->stream[0];
	dest->pub.free_in_buffer = STREAM_SIZE;

	dest->stream_count = 1;
}

METHODDEF(boolean)
empty_output_buffer (j_compress_ptr cinfo)
{
	mlibjpg_dest_ptr dest = (mlibjpg_dest_ptr) cinfo->dest;

	dest->stream = (void **)realloc(dest->stream, (dest->stream_count+1) * sizeof(void*));
	dest->stream[dest->stream_count] = malloc(STREAM_SIZE);

	dest->pub.next_output_byte = (unsigned char *)dest->stream[dest->stream_count];
	dest->pub.free_in_buffer = STREAM_SIZE;

	dest->stream_count++;

	return TRUE;
}

METHODDEF(void)
term_destination (j_compress_ptr cinfo)
{
	mlibjpg_dest_ptr dest = (mlibjpg_dest_ptr) cinfo->dest;
	//size_t datacount = OUTPUT_BUF_SIZE - dest->pub.free_in_buffer;
	dest->totalsize = STREAM_SIZE * (dest->stream_count - 1)
		+ (STREAM_SIZE - dest->pub.free_in_buffer);
}

void jpeg_memory_dest(j_compress_ptr cinfo)
{
	mlibjpg_dest_ptr dest;

	if (cinfo->dest == NULL) {	/* first time for this JPEG object? */
		cinfo->dest = (struct jpeg_destination_mgr *)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				SIZEOF(mlibjpg_destination_mgr));
	}

	dest = (mlibjpg_dest_ptr) cinfo->dest;
	dest->pub.init_destination = init_destination;
	dest->pub.empty_output_buffer = empty_output_buffer;
	dest->pub.term_destination = term_destination;
}

extern "C" MLIBJPG_API int SaveJPGToMemory(
		SaveJPGInfo *info,
		void *(*memory_callback)(int required_size, void *data),
		void *(*write_callback)(int line_number, void *data),
		void *data)
{
	jpeg_compress_struct cinfo;
	my_error_mgr jerr;
	JSAMPARRAY buffer;
	int row_stride;
	int result = 1;
	unsigned char *rp, *wp;
	int x,y;
	void *memory;
	int totalsize;

	if(memory_callback == NULL || write_callback == NULL)
		return 0;

try {

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	jpeg_create_compress(&cinfo);

	jpeg_memory_dest(&cinfo);

	cinfo.image_width = info->width; 	/* image width and height, in pixels */
	cinfo.image_height = info->height;
	cinfo.input_components = 3;		/* # of color components per pixel */
	cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
	jpeg_set_defaults(&cinfo);

	if(info->quality > 0)
		jpeg_set_quality(&cinfo, info->quality, info->force_baseline);

	if(info->progressive)
		jpeg_simple_progression(&cinfo);

	/* Now that we know input colorspace, fix colorspace-dependent defaults */
	//jpeg_default_colorspace(&cinfo);

	/* Start compressor */
	jpeg_start_compress(&cinfo, TRUE);

	/* Process data */
	row_stride = cinfo.image_width * cinfo.input_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
	for(y=0; y<info->height; y++)
	{
		rp = (unsigned char *)write_callback(y, data);
		if(rp == NULL) break;
		wp = buffer[0];
		for(x=0; x<info->width; x++,rp+=3){
			*wp++ = rp[2];
			*wp++ = rp[1];
			*wp++ = rp[0];
		}
		jpeg_write_scanlines(&cinfo, buffer, 1);
	}

	/* Finish compression and release memory */
	jpeg_finish_compress(&cinfo);

	totalsize = ((mlibjpg_dest_ptr)cinfo.dest)->totalsize;
	info->out_buffer_size = totalsize;

	memory = memory_callback(totalsize, data);
	for(x=0; totalsize>=0; x++,totalsize-=STREAM_SIZE) {
		memcpy(memory, ((mlibjpg_dest_ptr)cinfo.dest)->stream[x], 
			min(totalsize, STREAM_SIZE));
		free(((mlibjpg_dest_ptr)cinfo.dest)->stream[x]);
	}
	free(((mlibjpg_dest_ptr)cinfo.dest)->stream);

	jpeg_destroy_compress(&cinfo);

} catch(int /*error_code*/){
	result = 0;
}

	return result;
}	//SaveJPGToMemory()

