//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     DibIff.cpp
//       LoadIFF()
//       SaveIFF()
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "Dib.h"


// for iff_bmhd::masking
#define mskNone                 0
#define mskHasMask              1
#define mskHasTransparentColor	2
#define mskLasso	            3

// for iff_bmhd::compress
#define cmpNone                 0
#define cmpByteRun1	            1



class iff_chunk {
private:
	unsigned char data[8];
	long m_size;
public:

	bool load(FILE *fh) {
		if( fread(data, 8, 1, fh) != 1)
			return false;
		m_size = (long)(
			(data[4] << 24) | (data[5] << 16) | (data[6] << 8) | data[7]);
		return true;
	}
	inline bool check(const char *str) { return (memcmp(data, str, 4) == 0); }
	inline long size() const { return m_size; }
};


struct iff_bmhd
{
	unsigned short width;
	unsigned short height;
	unsigned short x_pos;
	unsigned short y_pos;
	unsigned char bpp;
	unsigned char mask;
	unsigned char compress;
	unsigned char pad;
	unsigned short transparent;
	unsigned char x_aspect;
	unsigned char y_aspect;
	unsigned short page_width;
	unsigned short page_height;

	iff_bmhd() { memset(this, 0, sizeof(*this)); }
};


inline unsigned char readChar(FILE *fh)
{
	return (unsigned char )fgetc(fh);
}

inline unsigned short readWord(FILE *fh)
{
	unsigned short w;
	w  = (unsigned short)(fgetc(fh) << 8);
	w |= (unsigned short)fgetc(fh);
	return w;
}

class iff_outstream
{
private:
	char *buffer;
	int max_size;
	int cur_size;

	void add(int len) {
		if(cur_size + len > max_size)
		{
			max_size = cur_size*2 + len + 256;
			buffer = (char *)realloc(buffer, max_size);
		}
	}
public:
	iff_outstream()
	{
		buffer = NULL;
		max_size = cur_size = 0;
	}

	~iff_outstream()
	{
		free(buffer);
	}

	inline int size() const { return cur_size; }
	inline char *ptr() const { return buffer; }

	void byte(char ch) {
		add(1);
		buffer[cur_size++] = ch;
	}
	void word(short ch) {
		add(2);
		buffer[cur_size++] = (char)((ch & 0xFF00) >> 8);
		buffer[cur_size++] = (char)(ch & 0xFF);
	}
	void dword(long ch) {
		add(4);
		buffer[cur_size++] = (char)((ch & 0xFF000000) >> 24);
		buffer[cur_size++] = (char)((ch & 0xFF0000) >> 16);
		buffer[cur_size++] = (char)((ch & 0xFF00) >> 8);
		buffer[cur_size++] = (char)(ch & 0xFF);
	}
	void dword(const char *ch) {
		add(4);
		memcpy(&buffer[cur_size], ch, 4);
		cur_size += 4;
	}
	void add(const void *ch, int len) {
		add(len);
		memcpy(&buffer[cur_size], ch, len);
		cur_size += len;
	}
};


static void LoadIFFImage(
	FILE *fh, iff_bmhd& header, DibStream *Dib, bool callback(float,void*), void *data)
{
	unsigned char *ptr;
	int i,j,k,x=0;
	int pitch = ((header.width + 15) / 16) * 2;
	int dib_pitch = Dib->GetPitch();
	int dib_bpp = Dib->GetBpp();
	bool eof = false;
	char rest[128];

	// allocate buffers
	char **row = (char **)malloc(sizeof(char *) * header.bpp);
	for(k=0; k<header.bpp; k++)
		row[k] = (char *)malloc(sizeof(char) * pitch + 128);

	for(j=0; j<header.height; j++)
	{
		// load an image per plane
		if(header.compress == 0)
		{
			for(k=0; k<header.bpp; k++)
				if(fread(row[k], pitch, 1, fh) != 1) { eof=true; break; }
		}
		else
		{
			for(k=0; k<header.bpp; k++)
			{
				if(x > 0)
					memcpy(row[k], rest, x);

				for(; x<pitch; )
				{
					int c = fgetc(fh);
					if(c>=0 && c<=127)
					{
						if(fread(&row[k][x], c+1, 1, fh) != 1) {eof=true; break; }
						x += c+1;
					}
					else if(c>=129 && c<=255)
					{
						int length = 257 - c;
						c = readChar(fh);
						memset(&row[k][x], c, length);
						x += length;
					}
					else if(c == EOF) {
						eof = true;
						break;
					}
				}
				x -= pitch;
				if(x > 0)
					memcpy(rest, &row[k][pitch], x);
			}
		}

		if(eof) break;

		unsigned char mask = 0x80;
		int x = 0;
		unsigned long col;
		ptr = (unsigned char *)Dib->ScanLine(j);

#define EXTRACT_COLOR \
				col = 0;							\
				for(k=header.bpp-1; k>=0; k--) {	\
					col <<= 1;						\
					if(row[k][x] & mask)			\
						col |= 1;					\
				}									\
				if((mask>>=1) == 0) { mask=0x80; x++; }

		switch(dib_bpp) {
		case 24:
			for(i=0; i<header.width; i++)
			{
				EXTRACT_COLOR;

				ptr[0] = (unsigned char)((col >> 16) & 0xFF); //B
				ptr[1] = (unsigned char)((col >>  8) & 0xFF); //G
				ptr[2] = (unsigned char)((col      ) & 0xFF); //R
				ptr += 3;
			}
			break;

		case 8:
			for(i=0; i<header.width; i++)
			{
				EXTRACT_COLOR;

				*ptr++ = (unsigned char)col;
			}
			break;

		case 4:
			memset(ptr, 0, dib_pitch);
			for(i=0; i<header.width; i++)
			{
				EXTRACT_COLOR;

				if(i & 1) {
					if(col & 0xF) *ptr |= 0xF;
					ptr++;
				} else {
					if(col & 0xF) *ptr |= 0xF0;
				}
			}
			break;

		case 1:
			memset(ptr, 0, dib_pitch);
			for(i=0; i<header.width; i++)
			{
				EXTRACT_COLOR;

				if(col) *ptr |= mask;
				if(mask == 0x80) ptr++;
			}
			break;
		}

		if(callback != NULL) {
			if( !callback((float)j/header.height,data) )
				break;
		}
	}

	for(k=0; k<header.bpp; k++)
		free(row[k]);
	free(row);
}


bool LoadIFF(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	bool result = false;
	bool initialized = false;
	long file_size, file_pos;
	iff_bmhd header;

	// open a file
	FILE *fh = fopen(FileName, "rb");
	if(fh == NULL)
		return false;

	// get a size of the file
	fseek(fh, 0, SEEK_END);
	file_size = ftell(fh);
	fseek(fh, 0, SEEK_SET);

	while((file_pos = ftell(fh)) < file_size)
	{
		iff_chunk chunk;

		if( !chunk.load(fh) )
			break;

		if(chunk.check("FORM"))
		{
			long sub_pos;
			char ilbm[4];

			// Is this ILBM chunk?
			if(fread(ilbm, 4, 1, fh) != 1) break;
			if(memcmp(ilbm, "ILBM", 4) != 0) break;

			// load sub-chunks
			while((sub_pos = ftell(fh)) < chunk.size())
			{
				iff_chunk subchunk;
				if( !subchunk.load(fh) )
					break;

				if(subchunk.check("BMHD"))
				{
					header.width    = readWord(fh);
					header.height   = readWord(fh);
					header.x_pos    = readWord(fh);
					header.y_pos    = readWord(fh);
					header.bpp      = readChar(fh);
					header.mask     = readChar(fh);
					header.compress = readChar(fh);
					header.pad      = readChar(fh);
					header.transparent = readWord(fh);
					header.x_aspect    = readChar(fh);
					header.y_aspect    = readChar(fh);
					header.page_width  = readWord(fh);
					header.page_height = readWord(fh);

					if(header.bpp <= 8)
					{
						if(header.bpp == 1)
							Dib->Create(header.width, header.height, 1);
						else if(header.bpp <= 4)
							Dib->Create(header.width, header.height, 4);
						else
							Dib->Create(header.width, header.height, 8);
						initialized = true;
					}
					else if(header.bpp <= 24)
					{
						Dib->Create(header.width, header.height, 24);
						initialized = true;
					}
				}
				else if(subchunk.check("CMAP"))
				{
					if(Dib->GetHeader() != NULL)
					{
						// read palettes
						int palettes = min((int)subchunk.size() / 3, Dib->GetNumColors());
						RGBQUAD *rgb = Dib->GetPalette();
						for(int i=0; i<palettes; i++,rgb++)
						{
							rgb->rgbRed      = readChar(fh);
							rgb->rgbGreen    = readChar(fh);
							rgb->rgbBlue     = readChar(fh);
							rgb->rgbReserved = 0;
						}
					}
				}
				else if(subchunk.check("BODY"))
				{
					if(initialized) {
						LoadIFFImage(fh, header, Dib, callback, data);
						result = true;
						break;
					}
				}

				// move to the next chunk
				fseek(fh, sub_pos+subchunk.size()+8, SEEK_SET);
			}
		}

		if(result)
			break;

		// move to the next chunk
		fseek(fh, file_pos+chunk.size()+8, SEEK_SET);
	}

	// close a file
	fclose(fh);

	return result;
}



bool SaveIFF(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	iff_outstream root, main_buf, img_buf;
	int width  = Dib->GetWidth();
	int height = Dib->GetHeight();
	int bpp    = Dib->GetBpp();
	int cmaps  = 0;
	DibStream *target;
	int i,j,k,n,x;
	bool result = false;

	main_buf.dword("ILBM");

	switch(bpp) {
	case 1:
	case 4:
	case 8:
	case 24:
		target = Dib;
		break;
	default:
		target = Dib->Clone();
		target->ChangeBpp(bpp = 24);
		break;
	}


	// header
	main_buf.dword("BMHD");
	main_buf.dword(20);
	main_buf.word((short)width);  // width
	main_buf.word((short)height); // height
	main_buf.word(0);             // x pos
	main_buf.word(0);             // y pos
	main_buf.byte((char)bpp);     // planes
	main_buf.byte(0);             // masking
	main_buf.byte(0);             // compress
	main_buf.byte(0);             // pad0 (not used)
	main_buf.word(0);             // transparent
	main_buf.byte(1);             // x aspect
	main_buf.byte(1);             // y aspect
	main_buf.word((short)width);  // page width
	main_buf.word((short)height); // page height

	// color palettes
	switch(bpp) {
	case 1: cmaps =   2; break;
	case 4: cmaps =  16; break;
	case 8: cmaps = 256; break;
	}

	if(cmaps > 0)
	{
		n = target->GetNumColors();
		main_buf.dword("CMAP");
		main_buf.dword(cmaps * 3);
		RGBQUAD *rgb = target->GetPalette();
		for(i=0; i<n; i++,rgb++) {
			main_buf.byte(rgb->rgbRed);
			main_buf.byte(rgb->rgbGreen);
			main_buf.byte(rgb->rgbBlue);
		}
		for(; i<cmaps; i++) {
			main_buf.byte(0);
			main_buf.byte(0);
			main_buf.byte(0);
		}
	}

	int pitch = ((width + 15) / 16) * 2;
	char **row = (char **)malloc(sizeof(char *) * bpp);
	for(k=0; k<bpp; k++)
		row[k] = (char *)malloc(sizeof(char) * pitch);

#define BIT_COLOR								\
		for(k=0; k<bpp; k++) {					\
			if(col & 1)							\
				row[k][x] |= mask;				\
			col >>= 1;							\
		}										\
		if((mask>>=1) == 0) { mask=0x80; x++; }	\

	unsigned char *ptr;
	unsigned char mask;
	unsigned long col;

	for(j=0; j<height; j++)
	{
		for(k=0; k<bpp; k++)
			memset(row[k], 0, pitch);

		ptr = (unsigned char *)target->ScanLine(j);
		mask = 0x80;

		switch(bpp) {
		case 1:
			for(i=x=0; i<width; i++)
			{
				col = (unsigned long)((*ptr & mask) ? 1 : 0);
				if(mask == 1) ptr++;

				BIT_COLOR;
			}
			break;

		case 4:
			for(i=x=0; i<width; i++)
			{
				if(i & 1) col = (unsigned long)(*ptr++ & 0xF);
				else      col = (unsigned long)((*ptr >> 4) & 0xF);

				BIT_COLOR;
			}
			break;

		case 8:
			for(i=x=0; i<width; i++)
			{
				col = (unsigned long)*ptr++;

				BIT_COLOR;
			}
			break;

		case 24:
			for(i=x=0; i<width; i++)
			{
				col = (unsigned long)((ptr[0] << 16) | (ptr[1] << 8) | ptr[2]);
				ptr += 3;

				BIT_COLOR;
			}
			break;
		}

		for(k=0; k<bpp; k++)
			img_buf.add(row[k], pitch);
	}

	// free memories
	if(target != Dib)
		delete target;
	for(k=0; k<bpp; k++)
		free(row[k]);
	free(row);

	main_buf.dword("BODY");
	main_buf.dword(img_buf.size());
	main_buf.add(img_buf.ptr(), img_buf.size());
	if(img_buf.size() & 1)
		main_buf.byte(0); // must be padded to an even length

	// open a file
	FILE *fh = fopen(FileName, "wb");
	if(fh != NULL)
	{
		// write an image
		root.dword("FORM");
		root.dword(main_buf.size());
		fwrite(root.ptr(), 8, 1, fh);
		fwrite(main_buf.ptr(), main_buf.size(), 1, fh);

		// close a file
		fclose(fh);

		result = true;
	}

	return result;
}




