//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     DibPpm.cpp
//       LoadPPM(), LoadPGM()
//       SavePPM(), SavePGM()
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "Dib.h"


static const char p1_header[]="P1\n"; // .PBM ASCII   1bit
static const char p2_header[]="P2\n"; // .PGM ASCII   8bit
static const char p3_header[]="P3\n"; // .PPM ASCII  24bit
static const char p4_header[]="P4\n"; // .PBM BINARY  1bit
static const char p5_header[]="P5\n"; // .PGM BINARY  8bit
static const char p6_header[]="P6\n"; // .PPM BINARY 24bit


//---------------------------------------------------------------------------
//   LoadPPM
//---------------------------------------------------------------------------
static int readInt(FILE *fh, int& result)
{
	bool begin = false;
	bool minus = false;
	int sum = 0;

	if(feof(fh))
		return false;

	while(1) {
		int ch = fgetc(fh);
		if(ch >= '0' && ch <= '9') {
			sum = sum * 10 + ch - '0';
			begin = true;
		} else if(ch == '-')
			minus = true;
		else if(begin)
			break;
		else if(ch == EOF)
			return false;
	}

	if(minus) sum = -sum;
	result = sum;

	return true;
}

bool LoadPPM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	int width, height, x, y, ppmpitch;
	char buf[128];
	bool result = false;
	int bpp;
	bool binary;
	BYTE *ptr, *tmpptr;
	int ch;

	FILE *fh = fopen(FileName,"rb");
	if(fh == NULL) return false;

	// Is it a PPM or PGM file?
	fgets(buf, sizeof(buf), fh);
	if(strcmp(buf, p6_header) == 0)      { binary=true;  bpp=24; }
	else if(strcmp(buf, p3_header) == 0) { binary=false; bpp=24; }
	else if(strcmp(buf, p5_header) == 0) { binary=true;  bpp= 8; }
	else if(strcmp(buf, p2_header) == 0) { binary=false; bpp= 8; }
	else {
		fclose(fh);
		return false;
	}

	// Skip any comments
	do {
		fgets(buf, sizeof(buf), fh);
	} while (buf[0] == '#');

	// Read the width and height
	sscanf(buf, "%d %d\n", &width, &height);
	fgets(buf, sizeof(buf), fh); //skip next line

	if( !Dib->Create(width, height, bpp) )
		return false;

	switch(bpp) {
	case 24:
		ppmpitch = width * 3;
		break;
	case  8:
		ppmpitch = width;
		Dib->SetMonoPalette();
		break;
	}

	BYTE *tmpbuf = (BYTE *)malloc(ppmpitch);

	// load an image
	switch(bpp) {
	case 24:
		for(y=0; y<height; y++)
		{
			ptr = (BYTE *)Dib->ScanLine(y);
			if(binary)
			{
				if(fread(tmpbuf, ppmpitch, 1, fh) != 1)
					break;
				tmpptr = tmpbuf;
				for(x=0; x<width; x++,ptr+=3,tmpptr+=3) {
					ptr[0] = tmpptr[2]; // B
					ptr[1] = tmpptr[1]; // G
					ptr[2] = tmpptr[0]; // R
				}
			}
			else
			{
				for(x=0; x<width; x++,ptr+=3)
				{
					if( !readInt(fh, ch) ) goto FILE_EXIT;
					ptr[2] = (unsigned char)ch; // R
					if( !readInt(fh, ch) ) goto FILE_EXIT;
					ptr[1] = (unsigned char)ch; // G
					if( !readInt(fh, ch) ) goto FILE_EXIT;
					ptr[0] = (unsigned char)ch; // B
				}
			}

			if(callback != NULL) {
				if( !callback((float)y/height, data) )
					break;
			}
		}
		break;

	case 8:
		for(y=0; y<height; y++)
		{
			ptr = (BYTE *)Dib->ScanLine(y);
			if(binary)
			{
				if(fread(ptr, ppmpitch, 1, fh) != 1)
					break;
			}
			else
			{
				for(x=0; x<width; x++,ptr++)
				{
					if( !readInt(fh, ch) ) goto FILE_EXIT;
					ptr[0] = (unsigned char)ch;
				}
			}

			if(callback != NULL) {
				if( !callback((float)y/height, data) )
					break;
			}
		}
		break;
	}

	if(y == height)
		result = true;

FILE_EXIT:
	free(tmpbuf);
	fclose(fh);

	return result;
}	//LoadPPM()


//---------------------------------------------------------------------------
//   LoadPGM
//---------------------------------------------------------------------------
bool LoadPGM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	return LoadPPM(FileName, Dib, callback, data);
}


//---------------------------------------------------------------------------
//   SavePPM
//---------------------------------------------------------------------------
bool SavePPM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	int width,height,x,y;
	DibStream *TmpDib;
	bool result = false;

	// create file
	FILE *fh = fopen(FileName,"wb");
	if(fh == NULL) return false;

	// create clone of DIB
	TmpDib = Dib->Clone();
	if(TmpDib->GetBpp() != 24)
		TmpDib->ChangeBpp(24);

	width  = TmpDib->GetWidth();
	height = TmpDib->GetHeight();

	// write a header
	fwrite(p6_header, 3, 1, fh);

	// write a size of the image
	fprintf(fh, "%d %d\n", width, height);

	// write ???
	fprintf(fh, "255\n");

	// write lines
	int ppmpitch = width * 3;
	BYTE *buf = (BYTE *)malloc(ppmpitch);
	for(y=0; y<height; y++)
	{
		BYTE *ptr = buf;
		BYTE *src = (BYTE *)TmpDib->ScanLine(y);
		for(x=0; x<width; x++,src+=3) {
			*ptr++ = src[2]; //r
			*ptr++ = src[1]; //g
			*ptr++ = src[0]; //b
		}
		fwrite(buf, ppmpitch, 1, fh);

		if(callback != NULL) {
			if( !callback((float)y/height,data) )
				goto FILE_EXIT;
		}
	}

	result = true;

FILE_EXIT:
	free(buf);

	// delete clone of DIB and close file
	delete TmpDib;
	fclose(fh);

	return result;
}	//SavePPM()


//---------------------------------------------------------------------------
//   SavePGM
//---------------------------------------------------------------------------
bool SavePGM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	int width, height, y;
	DibStream *TmpDib;
	bool result = false;

	// create file
	FILE *fh = fopen(FileName, "wb");
	if(fh == NULL)
		return false;

	// create a clone of DIB
	TmpDib = Dib->Clone();
	TmpDib->ChangeBpp(8, DIB_CHANGEBPP_8_MONO);

	width  = TmpDib->GetWidth();
	height = TmpDib->GetHeight();

	// write a header
	fwrite(p5_header, 3, 1, fh);

	// write a size of the image
	fprintf(fh, "%d %d\n", width, height);

	// write ???
	fprintf(fh, "255\n");

	// write lines
	for(y=0; y<height; y++)
	{
		fwrite(TmpDib->ScanLine(y), width, 1, fh);
		if(callback != NULL) {
			if( !callback((float)y/height, data) )
				goto FILE_EXIT;
		}
	}

	result = true;

FILE_EXIT:
	// delete clone of DIB and close file
	delete TmpDib;
	fclose(fh);

	return result;
}	//SavePGM()



