//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     DibTga.cpp
//       LoadTGA()
//       SaveTGA()
//---------------------------------------------------------------------------

#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "Dib.h"

//---------------------------------------------------------------------------

struct TgaHeader
{
	void GetFromBuffer(char *buffer)
	{
		CharaNum  = (BYTE)buffer[0];
		ColorMap  = (BYTE)buffer[1];
		ImageType = (BYTE)buffer[2];
		ColorMapOrigin = *(short *)&buffer[3];
		ColorMapLength = *(short *)&buffer[5];
		ColorMapEntrySize = (BYTE)buffer[7];
		XOrigin = *(short *)&buffer[8];
		YOrigin = *(short *)&buffer[10];
		Width   = *(short *)&buffer[12];
		Height  = *(short *)&buffer[14];
		PixelSize  = (BYTE)buffer[16];
		Descriptor = (BYTE)buffer[17];
	}

	void SetToBuffer(char *buffer)
	{
		buffer[0] = (BYTE)CharaNum;
		buffer[1] = (BYTE)ColorMap;
		buffer[2] = (BYTE)ImageType;
		*(short *)&buffer[3] = ColorMapOrigin;
		*(short *)&buffer[5] = ColorMapLength;
		buffer[7] = (BYTE)ColorMapEntrySize;
		*(short *)&buffer[8] = XOrigin;
		*(short *)&buffer[10] = YOrigin;
		*(short *)&buffer[12] = Width;
		*(short *)&buffer[14] = Height;
		buffer[16] = (BYTE)PixelSize;
		buffer[17] = (BYTE)Descriptor;
	}

	BYTE CharaNum; // Number of Characters in Identification Field.
	BYTE ColorMap; // 0:no color map , 1:a color map is included
	BYTE ImageType;
	short ColorMapOrigin; // Integer ( lo-hi ) index of first color map entry
	short ColorMapLength; // Integer ( lo-hi ) count of color map entries
	BYTE ColorMapEntrySize; // Number of bits in color map entry  16,24,32
	short XOrigin;
	short YOrigin;
	short Width;
	short Height;
	BYTE PixelSize; //Number of bits in a stored pixel index  16,24,....
	BYTE Descriptor;
};


// TgaHeader::Descriptor
//  Bits 3-0 - number of attribute bits associated with each pixel.
//  Bit 4    - reserved.  Must be set to 0.
//  Bit 5    - screen origin bit.
//             0 = Origin in lower left-hand corner.
//             1 = Origin in upper left-hand corner.
//             Must be 0 for Truevision images.
//  Bits 7-6 - Data storage interleaving flag.
//             00 = non-interleaved.
//             01 = two-way (even/odd) interleaving.
//             10 = four way interleaving.
//             11 = reserved.

#define TGA_HEADER_SIZE  18

// TgaHeader::ImageType
#define TGA_TYPE_NOIMAGE                       0
#define TGA_TYPE_COLORMAPPED                   1
#define TGA_TYPE_RGB                           2
#define TGA_TYPE_MONO                          3
#define TGA_TYPE_RLE_COLORMAPPED               9
#define TGA_TYPE_RLE_RGB                      10
#define TGA_TYPE_COMPRESSED_MONO              11
#define TGA_TYPE_COMPRESSED_COLORMAPPED       32
#define TGA_TYPE_COMPRESSED_COLORMAPPED_QUAD  33

// TgaHeader::Descriptor (bit0-3)
#define TGA_DESCRIPTOR_RGB_16     0
#define TGA_DESCRIPTOR_RGB_24     0
#define TGA_DESCRIPTOR_RGB_32     8
// TgaHeader::Descriptor (bit5)
#define TGA_DESCRIPTOR_LOWER      0
#define TGA_DESCRIPTOR_UPPER   0x20



static int GetByteBpp(int Bpp)
{
	switch(Bpp) {
	case  8: return 1;
	case 16: return 2;
	case 24: return 3;
	case 32: return 4;
	}
	return 0;
}


static bool LoadRLE_COLORMAPPED(int fh, DibStream *Dib, TgaHeader& tga, bool callback(float,void*), void *data)
{
	char ch;
	BYTE col;
	BYTE *ptr;
	int x,y,dy,ey,length;

	// check bpp
	switch(tga.PixelSize) {
	case 8: break;
	default: return false;
	}

	// check if it has a color map
	if(!tga.ColorMap) return false;

	// create a DIB
	if( !Dib->Create(tga.Width, tga.Height, 8) )
		return false;
	if(tga.Descriptor & TGA_DESCRIPTOR_UPPER) {
		y = 0;
		dy = 1;
		ey = tga.Height;
	} else {
		y = tga.Height - 1;
		dy = -1;
		ey = -1;
	}

	// set palettes
	switch(tga.ColorMapEntrySize) {
	case 24:
		{
			BYTE *buf = (BYTE *)malloc(3 * tga.ColorMapLength);
			if( read(fh, buf, 3*tga.ColorMapLength) != 3*tga.ColorMapLength)
				{ free(buf); return false; }
			RGBQUAD *rgb = Dib->GetColors();
			for(x=0,ptr=buf; x<tga.ColorMapLength; x++,rgb++) {
				rgb->rgbBlue  = *ptr++;
				rgb->rgbGreen = *ptr++;
				rgb->rgbRed   = *ptr++;
				rgb->rgbReserved = 0;
			}
			free(buf);
		}
		break;

	default:
		return false;
	}

	// set an image
	ptr = (BYTE *)Dib->ScanLine(y);
	for(x=0; y!=ey; )
	{
		if( read(fh, &ch, 1) != 1) return false;

		length = (ch & 0x7F) + 1;
		if(ch & 0x80)
		{
			// the run-length packet
			if( read(fh, &col, 1) != 1) return false;
			for(; length>0; length--) {
				*ptr++ = col;
				if(++x >= tga.Width) {
					x = 0;
					y += dy;
					if(y == ey) break;
					ptr = (BYTE *)Dib->ScanLine(y);
				}
			}
		}
		else
		{
			// the raw packet
			for(; length>0; length--) {
				if( read(fh, &col, 1) != 1) return false;
				*ptr++ = col;
				if(++x >= tga.Width) {
					x = 0;
					y += dy;
					if(y == ey) break;
					ptr = (BYTE *)Dib->ScanLine(y);
				}
			}
		}
	}

	// succeeded!
	return true;
}


static bool LoadRLE_RGB(int fh, DibStream *Dib, TgaHeader& tga, bool callback(float,void*), void *data)
{
	char ch;
	union {
		BYTE col[4];
		WORD wcol[2];
		DWORD dwcol;
	};
	BYTE *ptr;
	int x,y,dy,ey,length;

	// check bpp
	switch(tga.PixelSize) {
	case 16: break;
	case 24: break;
	case 32: break;
	default: return false;
	}

	// create a DIB
	if( !Dib->Create(tga.Width, tga.Height, tga.PixelSize) )
		return false;
	if(tga.Descriptor & TGA_DESCRIPTOR_UPPER) {
		y = 0;
		dy = 1;
		ey = tga.Height;
	} else {
		y = tga.Height - 1;
		dy = -1;
		ey = -1;
	}

	// ignore palettes
	if(tga.ColorMap) {
		lseek(fh,
			GetByteBpp(tga.ColorMapEntrySize) * tga.ColorMapLength,
			SEEK_CUR);
	}

	// set an image
	ptr = (BYTE *)Dib->ScanLine(y);
	x = 0;
	switch(tga.PixelSize) {
	case 24:
		for(; y!=ey; )
		{
			if( read(fh, &ch, 1) != 1) return false;

			length = (ch & 0x7F) + 1;
			if(ch & 0x80)
			{
				// the run-length packet
				if( read(fh, &col, 3) != 3) return false;
				for(; length>0; length--) {
					ptr[0] = col[0];
					ptr[1] = col[1];
					ptr[2] = col[2];
					ptr += 3;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
			else
			{
				// the raw packet
				for(; length>0; length--) {
					if( read(fh, &col, 3) != 3) return false;
					ptr[0] = col[0];
					ptr[1] = col[1];
					ptr[2] = col[2];
					ptr += 3;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
		}
		break;

	case 32:
		for(; y!=ey; )
		{
			if( read(fh, &ch, 1) != 1) return false;

			length = (ch & 0x7F) + 1;
			if(ch & 0x80)
			{
				// the run-length packet
				if( read(fh, &col, 4) != 4) return false;
				for(; length>0; length--) {
					*(DWORD *)ptr = dwcol;
					ptr += 4;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
			else
			{
				// the raw packet
				for(; length>0; length--) {
					if( read(fh, &col, 4) != 4) return false;
					*(DWORD *)ptr = dwcol;
					ptr += 4;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
		}
		break;

	case 16:
		for(; y!=ey; )
		{
			if( read(fh, &ch, 1) != 1) return false;

			length = (ch & 0x7F) + 1;
			if(ch & 0x80)
			{
				// the run-length packet
				if( read(fh, &col, 2) != 2) return false;
				for(; length>0; length--) {
					*(WORD *)ptr = wcol[0];
					ptr += 2;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
			else
			{
				// the raw packet
				for(; length>0; length--) {
					if( read(fh, &col, 2) != 2) return false;
					*(WORD *)ptr = wcol[0];
					ptr += 2;
					if(++x >= tga.Width) {
						x = 0;
						y += dy;
						if(y == ey) break;
						ptr = (BYTE *)Dib->ScanLine(y);
					}
				}
			}
		}
		break;
	}

	// invert the alpha channel
	if(tga.PixelSize == 32)
	{
		for(y=0; y<tga.Height; y++) {
			ptr = (BYTE *)Dib->ScanLine(y) + 3;
			for(x=0; x<tga.Width; x++,ptr+=4)
				*ptr = (BYTE)(255 - *ptr);
		}
	}

	// succeeded!
	return true;
}


//---------------------------------------------------------------------------
//   LoadTGA
//---------------------------------------------------------------------------
bool LoadTGA(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	TgaHeader tga;
	int i,x,y,dibpitch,tgapitch;
	LPBYTE buf,ptr;
	RGBQUAD *rgb;
	char header[TGA_HEADER_SIZE];
	bool result = false;

	// open a file
	int fh = open(FileName, O_RDONLY|O_BINARY);
	if(fh == -1) return false;

	// load a header
	if( read(fh, header, TGA_HEADER_SIZE) != TGA_HEADER_SIZE )
		goto LOAD_EXIT;
	tga.GetFromBuffer(header);

	// ignore 'Image Identification Field'
	if(tga.CharaNum)
		lseek(fh, tga.CharaNum, SEEK_CUR);

	switch(tga.ImageType) {
	case TGA_TYPE_COLORMAPPED:
		// check bpp
		switch(tga.PixelSize) {
		case 8: tgapitch=tga.Width; break;
		default: goto LOAD_EXIT;
		}
		// create a DIB
		if(tga.Descriptor & TGA_DESCRIPTOR_UPPER){
			if( !Dib->Create(tga.Width, -tga.Height, 8) )
				goto LOAD_EXIT;
		} else {
			if( !Dib->Create(tga.Width, tga.Height, 8) )
				goto LOAD_EXIT;
		}
		// set palettes
		if(!tga.ColorMap) goto LOAD_EXIT;
		switch(tga.ColorMapEntrySize) {
		case 24:
			buf = (BYTE *)malloc(3 * tga.ColorMapLength);
			if( read(fh, buf, 3*tga.ColorMapLength) != 3*tga.ColorMapLength)
				{ free(buf); goto LOAD_EXIT; }
			rgb = Dib->GetColors();
			for(i=0,ptr=buf; i<tga.ColorMapLength; i++) {
				rgb[i].rgbBlue = *ptr++;
				rgb[i].rgbGreen = *ptr++;
				rgb[i].rgbRed = *ptr++;
				rgb[i].rgbReserved = 0;
			}
			free(buf);
			break;
		default:
			goto LOAD_EXIT;
		}
		// set an image
		dibpitch = Dib->GetPitch();
		buf = (LPBYTE)Dib->GetBuffer();
		for(y=0; y<tga.Height; y++) {
			if( read(fh, &buf[dibpitch*y], tgapitch) != tgapitch)
				goto LOAD_EXIT;
		}
		// succeeded!
		result = true;
		break;

	case TGA_TYPE_RGB:
		// check bpp
		switch(tga.PixelSize) {
		case 16: tgapitch=2*tga.Width; break;
		case 24: tgapitch=3*tga.Width; break;
		case 32: tgapitch=4*tga.Width; break;
		default: goto LOAD_EXIT;
		}
		// create a DIB
		if(tga.Descriptor & TGA_DESCRIPTOR_UPPER) {
			if( !Dib->Create(tga.Width, -tga.Height, tga.PixelSize) )
				goto LOAD_EXIT;
		} else {
			if( !Dib->Create(tga.Width, tga.Height, tga.PixelSize) )
				goto LOAD_EXIT;
		}
		// ignore palettes
		if(tga.ColorMap) {
			lseek(fh,
				GetByteBpp(tga.ColorMapEntrySize) * tga.ColorMapLength,
				SEEK_CUR);
		}
		// set an image
		dibpitch = Dib->GetPitch();
		buf = (LPBYTE)Dib->GetBuffer();
		for(y=0; y<tga.Height; y++) {
			if( read(fh, &buf[dibpitch*y], tgapitch) != tgapitch)
				goto LOAD_EXIT;
		}
		// invert the alpha channel
		if(tga.PixelSize == 32)
		{
			for(y=0; y<tga.Height; y++) {
				LPBYTE ptr = &buf[dibpitch*y+3];
				for(x=0; x<tga.Width; x++,ptr+=4)
					*ptr = (BYTE)(255 - *ptr);
			}
		}
		// succeeded!
		result = true;
		break;

	case TGA_TYPE_MONO:
		// check bpp
		switch(tga.PixelSize) {
		case 8: tgapitch=tga.Width; break;
		default: goto LOAD_EXIT;
		}
		// create a DIB
		if(tga.Descriptor & TGA_DESCRIPTOR_UPPER) {
			if( !Dib->Create(tga.Width, -tga.Height, 8) )
				goto LOAD_EXIT;
		} else {
			if( !Dib->Create(tga.Width, tga.Height, 8) )
				goto LOAD_EXIT;
		}
		// set a palette
		Dib->SetMonoPalette();
		// ignore a palette
		if(tga.ColorMap) {
			lseek(fh,
				GetByteBpp(tga.ColorMapEntrySize) * tga.ColorMapLength,
				SEEK_CUR);
		}
		// set an image
		dibpitch = Dib->GetPitch();
		buf = (LPBYTE)Dib->GetBuffer();
		for(y=0; y<tga.Height; y++) {
			if( read(fh, &buf[dibpitch*y], tgapitch) != tgapitch)
				goto LOAD_EXIT;
		}
		// succeeded!
		result = true;
		break;

	case TGA_TYPE_RLE_COLORMAPPED:
		result = LoadRLE_COLORMAPPED(fh, Dib, tga, callback, data);
		break;
		
	case TGA_TYPE_RLE_RGB:
		result = LoadRLE_RGB(fh, Dib, tga, callback, data);
		break;
	}

LOAD_EXIT:
	// close a file
	close(fh);

	return result;
}	//LoadTGA()


//---------------------------------------------------------------------------
//   SaveTGA
//---------------------------------------------------------------------------
bool SaveTGA(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	TgaHeader tga;
	int w,h,x,y,pitch;
	char header[TGA_HEADER_SIZE];
	DibStream *TmpDib;
	int i;
	bool result = false;

	// create file
	int fh = open(FileName, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, S_IREAD|S_IWRITE);
	if(fh == -1) return false;

	// create clone of DIB
	TmpDib = Dib->Clone();

	// change bpp
	if(TmpDib->GetBpp() < 8)
		TmpDib->ChangeBpp(24);

	w = TmpDib->GetWidth();
	h = TmpDib->GetHeight();

	if(TmpDib->GetBpp() == 8)
	{
		// write header
		tga.CharaNum = 0;
		tga.ColorMap = 1;
		tga.ImageType = TGA_TYPE_COLORMAPPED;
		tga.ColorMapOrigin = 0; //Integer ( lo-hi ) index of first color map entry
		tga.ColorMapLength = (short)TmpDib->GetNumColors(); //Integer ( lo-hi ) count of color map entries
		tga.ColorMapEntrySize = 24; //Number of bits in color map entry  16,24,32
		tga.XOrigin = 0;
		tga.YOrigin = 0;
		tga.Width  = (short)w;
		tga.Height = (short)h;
		tga.PixelSize = (BYTE)TmpDib->GetBpp(); //Number of bits in a stored pixel index  16,24,....
		tga.Descriptor = 0;
		tga.SetToBuffer(header);
		write(fh, header, TGA_HEADER_SIZE);

		// write palette
		BYTE *pal = (BYTE *)malloc(3 * tga.ColorMapLength);
		RGBQUAD *rgb = Dib->GetColors();
		BYTE *ptr = pal;
		for(i=0; i<tga.ColorMapLength; i++){
			*ptr++ = rgb[i].rgbBlue;
			*ptr++ = rgb[i].rgbGreen;
			*ptr++ = rgb[i].rgbRed;
		}
		write(fh, pal, 3*tga.ColorMapLength);
		free(pal);

		// write lines (bottom-up)
		pitch = tga.Width;
		for(y=0; y<h; y++){
			write(fh, TmpDib->ScanLine(h-y-1), pitch);
			if(callback != NULL){
				if(!callback((float)y/h,data))
					goto FILE_EXIT;
			}
		}
	}
	else
	{
		// write header
		tga.CharaNum = 0;
		tga.ColorMap = 0; //no color map
		tga.ImageType = TGA_TYPE_RGB;
		tga.ColorMapOrigin = 0; //Integer ( lo-hi ) index of first color map entry
		tga.ColorMapLength = 0; //Integer ( lo-hi ) count of color map entries
		tga.ColorMapEntrySize = 0; //Number of bits in color map entry  16,24,32
		tga.XOrigin = 0;
		tga.YOrigin = 0;
		tga.Width  = (short)w;
		tga.Height = (short)h;
		tga.PixelSize = (BYTE)TmpDib->GetBpp(); //Number of bits in a stored pixel index  16,24,....
		tga.Descriptor = TGA_DESCRIPTOR_LOWER;
		switch(TmpDib->GetBpp()){
		case 16: tga.Descriptor |= TGA_DESCRIPTOR_RGB_16; break;
		case 24: tga.Descriptor |= TGA_DESCRIPTOR_RGB_24; break;
		case 32: tga.Descriptor |= TGA_DESCRIPTOR_RGB_32; break;
		}
		tga.SetToBuffer(header);
		write(fh, header, TGA_HEADER_SIZE);

		// invert alpha
		if(tga.PixelSize == 32)
		{
			for(y=0; y<h; y++){
				LPBYTE ptr = (LPBYTE)TmpDib->ScanLine(y) + 3;
				for(x=0; x<w; x++,ptr+=4)
					*ptr = (BYTE)(255 - *ptr);
			}
		}

		// write lines (bottom-up)
		switch(TmpDib->GetBpp()){
		case 16: pitch=2; break;
		case 24: pitch=3; break;
		case 32: pitch=4; break;
		default: pitch=0;
		}
		pitch *= tga.Width;
		for(y=0; y<h; y++){
			write(fh, TmpDib->ScanLine(h-y-1), pitch);
			if(callback != NULL){
				if(!callback((float)y/h,data))
					goto FILE_EXIT;
			}
		}
	}

	result = true;

FILE_EXIT:
	// delete clone of DIB and close file
	delete TmpDib;
	close(fh);

	return result;
}	//SaveTGA()

